package be.iminds.ilabt.jfed.experiment.setup;

import be.iminds.ilabt.jfed.espec.bundle.BundleFetcher;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.bundle.ModifiedESpecBundle;
import be.iminds.ilabt.jfed.espec.filefetcher.FileFetcher;
import be.iminds.ilabt.jfed.espec.model.BasicExperimentSpecificationBuilder;
import be.iminds.ilabt.jfed.espec.model.ExperimentSpecificationBuilder;
import be.iminds.ilabt.jfed.espec.model.FileSource;
import be.iminds.ilabt.jfed.espec.model.RspecSpec;
import be.iminds.ilabt.jfed.espec.parser.ExperimentSpecificationParser;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentControllerFactory;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.git.SingleSshGitAuthPreferences;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.TestLinksJob;
import be.iminds.ilabt.jfed.highlevel.jobs.states.JobStateFactory;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.ProxyServiceUtil;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.credential.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.finder.AuthorityFinder;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.rspec.adv_based_generator.AdvertisementBasedRspecGenerator;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGenerator;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGeneratorFactory;
import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecFactory;
import be.iminds.ilabt.jfed.rspec.model.RspecFactoryFactory;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.rspec.util.RspecBinder;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import be.iminds.ilabt.jfed.util.common.TextUtil;
import be.iminds.ilabt.jfed.util.library.KeyUtil;
import com.google.inject.Injector;
import org.jetbrains.annotations.Contract;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import static be.iminds.ilabt.jfed.espec.model.FileSource.SourceType.DIRECT;

public class ExperimentSetupHelper {
    @Nonnull private final String uniqueId;

    @Nonnull private final ExperimentSetupLogger logger;
    @Nonnull private final ExperimentSetupConfig config;
    @Nonnull private final ExperimentSetupContext context;

    @Nonnull private final Supplier<RSpecGeneratorFactory> rSpecGeneratorFactoryProvider;
    @Nonnull private final RspecFactory rspecFactory;
    @Nullable private final AdvertisementRspecSource advertisementRspecSource;
    @Nonnull private final TestbedInfoSource testbedInfoSource;
    @Nonnull private final AuthorityFinder authorityFinder;
    @Nonnull private final SfaModel sfaModel;
    @Nonnull private final GeniUser user;
    @Nonnull private final AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory auasawf;
    @Nonnull private final Logger callLogger;

    private boolean notEnoughFreeResourcesDetected;

    public ExperimentSetupHelper(@Nullable String uniqueId,

                                 @Nonnull ExperimentSetupLogger logger,
                                 @Nonnull ExperimentSetupConfig config,
                                 @Nullable ExperimentSetupContext context,

                                 @Nonnull Supplier<RSpecGeneratorFactory> rSpecGeneratorFactoryProvider,
                                 @Nonnull RspecFactory rspecFactory,
                                 @Nullable AdvertisementRspecSource advertisementRspecSource,
                                 @Nonnull TestbedInfoSource testbedInfoSource,
                                 @Nonnull AuthorityFinder authorityFinder,
                                 @Nonnull GeniUser user,
                                 AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory auasawf,
                                 @Nonnull SfaModel sfaModel,
                                 @Nonnull Logger callLogger) {
        this.uniqueId = uniqueId == null ? "" : uniqueId;

        this.logger = logger;
        this.config = config;
        this.context = context == null ? new ExperimentSetupContext() : context;

        this.rSpecGeneratorFactoryProvider = rSpecGeneratorFactoryProvider;
        this.rspecFactory = rspecFactory;
        this.advertisementRspecSource = advertisementRspecSource;
        this.testbedInfoSource = testbedInfoSource;
        this.authorityFinder = authorityFinder;
        this.sfaModel = sfaModel;
        this.user = user;
        this.auasawf = auasawf;
        this.callLogger = callLogger;

        this.notEnoughFreeResourcesDetected = false;
    }

    public ExperimentSetupHelper(@Nullable String uniqueId,

                                 @Nonnull ExperimentSetupLogger logger,
                                 @Nonnull ExperimentSetupConfig config,
                                 @Nullable ExperimentSetupContext context,

                                 @Nonnull Supplier<RSpecGeneratorFactory> rSpecGeneratorFactoryProvider,
                                 @Nullable AdvertisementRspecSource advertisementRspecSource,

                                 @Nonnull Injector injector) {
        this.uniqueId = uniqueId == null ? "" : uniqueId;

        this.logger = logger;
        this.config = config;
        this.context = context == null ? new ExperimentSetupContext() : context;

        this.rSpecGeneratorFactoryProvider = rSpecGeneratorFactoryProvider;
        this.advertisementRspecSource = advertisementRspecSource;

        this.rspecFactory = RspecFactoryFactory.getRspecFactoryInstance(ModelRspecType.BASIC);
        this.testbedInfoSource = injector.getInstance(TestbedInfoSource.class);
        this.authorityFinder = injector.getInstance(AuthorityFinder.class);
        this.sfaModel = injector.getInstance(SfaModel.class);
        this.user = injector.getInstance(GeniUser.class);
        this.auasawf = injector.getInstance(AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory.class);
        this.callLogger = injector.getInstance(Logger.class);

        this.notEnoughFreeResourcesDetected = false;
    }

    @Nonnull
    public ExperimentSetupLogger getLogger() {
        return logger;
    }

    @Nonnull
    public ExperimentSetupConfig getConfig() {
        return config;
    }

    @Nonnull
    public ExperimentSetupContext getContext() {
        return context;
    }

    public void getRspec(@Nullable String generateAndBindTargetCmUrn) throws GeniUrn.GeniUrnParseException,
            ESpecBundle.ESpecBundleInitException,
            InterruptedException,
            IOException,
            ExperimentSpecificationParser.ExperimentSpecificationParseException {
        try {
            ExperimentSetupConfig.RequestRSpec rspecConfig = config.getRequestRSpec();
            if (rspecConfig != null) {
                switch (rspecConfig.getSource()) {
                    case GENERATE_USING_ADVERTISEMENT: {
                        ExperimentSetupConfig.GenerateUsingAdvertisementRequestRSpec specRspecConfig =
                                (ExperimentSetupConfig.GenerateUsingAdvertisementRequestRSpec) rspecConfig;
                        assertNotNull(this.advertisementRspecSource, "internal error");
                        assertNotNull(specRspecConfig.getConfig(), "internal error: check not done yet");
                        AdvertisementBasedRspecGenerator gen =
                                new AdvertisementBasedRspecGenerator(
                                        rspecFactory,
                                        specRspecConfig.getConfig(),
                                        this.advertisementRspecSource,
                                        testbedInfoSource
                                );
                        context.requestRspec = gen.generate().getRspecXmlString();
                        logger.info("Generated Request RSpec using \"GENERATE_USING_ADVERTISEMENT\" method:");
                        logger.info(context.requestRspec, true);
                        break;
                    }
                    case GENERATE_USING_JFED: {
                        if (generateAndBindTargetCmUrn == null)
                            logger.fatal("GENERATE_USING_JFED requires a CM URN");
                        ExperimentSetupConfig.GenerateUsingJFedRequestRSpec specRspecConfig =
                                (ExperimentSetupConfig.GenerateUsingJFedRequestRSpec) rspecConfig;
                        RSpecGenerator gen = rSpecGeneratorFactoryProvider.get().createRSpecGenerator();
                        gen.setAm(generateAndBindTargetCmUrn);
                        gen.setNodeCount(specRspecConfig.getNodeCount());
                        gen.setResourceClassIdPrefix(specRspecConfig.getjFedResourceClass());
                        gen.setNodeAnsibleGroup(specRspecConfig.getNodeAnsibleGroup());
                        context.requestRspec = gen.generateRequest();
                        logger.info("Generated Request RSpec using \"GENERATE_USING_JFED\" method:");
                        logger.info(context.requestRspec, true);
                        break;
                    }
                    case GENERATE_SIMPLE: {
                        if (generateAndBindTargetCmUrn == null)
                            logger.fatal("GENERATE_SIMPLE requires a CM URN");
                        ExperimentSetupConfig.GenerateSimpleRequestRSpec specRspecConfig =
                                (ExperimentSetupConfig.GenerateSimpleRequestRSpec) rspecConfig;
                        @Nonnull String exclusive = specRspecConfig.getNodeExclusive() ? "true" : "false";
                        @Nonnull String sliverTypeName = specRspecConfig.getNodeSliverType();
                        int nodeCount = specRspecConfig.getNodeCount();
                        @Nonnull List<String> compIds = specRspecConfig.getNodeComponentUrn();
                        assertTrue(compIds.isEmpty() || nodeCount == compIds.size(), "Config error: Mismatch of componentIds and nodeCount");
                        @Nonnull String compManId = generateAndBindTargetCmUrn;
                        context.requestRspec = "<?xml version='1.0'?>\n" +
                                "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" type=\"request\"  " +
                                "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                                "xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \">\n";

                        for (int i = 0; i < nodeCount; i++) {
                            String ansibleGroup = i < specRspecConfig.getNodeAnsibleGroup().size() ?
                                    specRspecConfig.getNodeAnsibleGroup().get(i) :
                                    (specRspecConfig.getNodeAnsibleGroup().isEmpty() ?
                                            null :
                                            specRspecConfig.getNodeAnsibleGroup().get(specRspecConfig.getNodeAnsibleGroup().size() - 1)
                                    );

                            @Nonnull String compId = "";
                            if (compIds.size() > i)
                                compId = " component_id=\"" + compIds.get(i) + "\"";
                            context.requestRspec +=
                                    "  <node client_id=\"n" + i + "\" exclusive=\"" + exclusive + "\" component_manager_id=\"" + compManId + "\"" + compId + ">\n" +
                                            "    <sliver_type name=\"" + sliverTypeName + "\"/>\n" +
                                            (ansibleGroup == null ?
                                                    "" :
                                                    "    <ansible_group xmlns=\"http://jfed.iminds.be/rspec/ext/jfed/1\" name=\"" + ansibleGroup + "\"/>\n"
                                            ) +
                                            "  </node>\n";
                        }

                        context.requestRspec += "</rspec>";

                        logger.info("Generated Request RSpec using \"simple\" method:");
                        logger.info(context.requestRspec, true);
                        break;
                    }
                    case PROVIDE_URL: {
                        ExperimentSetupConfig.ProvidedRequestRSpec specRspecConfig =
                                (ExperimentSetupConfig.ProvidedRequestRSpec) rspecConfig;
                        String rspecUrl = specRspecConfig.getProvidedContentSource();
                        assertNotNull(rspecUrl, "Config error: no RSpec URL provided");

                        URL url;
                        try {
                            url = new URL(rspecUrl);
                        } catch (MalformedURLException e) {
                            throw new RuntimeException("Invalid URL specified for RSpec: \"" + rspecUrl + "\"", e);
                        }
                        try (InputStream is = url.openStream()) {
                            String rspec = IOUtils.streamToString(is, "UTF-8");
                            if (rspec.trim().isEmpty()) {
                                logger.warn("Empty Rspec read from URL \"" + rspecUrl + "\"");
                            }

                            String head = TextUtil.abbreviate(rspec, 200);
                            logger.info("Downloaded Rspec from \"" + rspecUrl + "\". Rspec size: " + rspec.length() + " characters.\nHead: " + head, true/*formatAsCode*/);
                            context.requestRspec = rspec;
                        } catch (IOException e) {
                            throw new RuntimeException("Failed to fetch RSpec at \"" + rspecUrl + "\"", e);
                        }
                        break;
                    }
                    case PROVIDE_CONTENT: {
                        ExperimentSetupConfig.ProvidedRequestRSpec specRspecConfig =
                                (ExperimentSetupConfig.ProvidedRequestRSpec) rspecConfig;
                        context.requestRspec = specRspecConfig.getProvidedContentSource();
                        assertNotNull(context.requestRspec, "Config error: no RSpec provided");

                        String head = TextUtil.abbreviate(context.requestRspec, 200);
                        logger.info("Using Request Rspec from config. " +
                                "Size: " + context.requestRspec.length() + " characters.\n" +
                                "Head: " + head, true/*formatAsCode*/);

                        break;
                    }
                    default:
                        throw new RuntimeException("RequestRSpec source " + config.getRequestRSpec().getSource() + " not yet implemented");
                }

                //noinspection ConstantConditions
                if (rspecConfig.getSource().isProvide() &&
                        ((ExperimentSetupConfig.ProvidedRequestRSpec) rspecConfig)
                                .getBindTestedServerToUnboundNodes()) {
                    String origRequestRspec = context.requestRspec;
                    context.requestRspec = RspecBinder.bindUnboundTo(
                            logger::warn, s -> logger.warn(s, true), logger::info,
                            context.requestRspec, generateAndBindTargetCmUrn
                    );
                    if (!origRequestRspec.equals(context.requestRspec)) {
                        logger.info("Unbound nodes were bound to " + generateAndBindTargetCmUrn);
                    }
                }

                logger.logExtraXml("rspec"+uniqueId, context.requestRspec, true);
                context.allRequestedNodeComponentIds.addAll(logRequestRspecFixedNodes(context.requestRspec));

                context.requestRspecSource = new RequestRspecSource(context.requestRspec, ModelRspecType.BASIC);
                context.serversInRspec = context.requestRspecSource.getAllComponentManagerUrns().stream()
                        .map(serverUrn ->
                                authorityFinder.findByUrn(
                                        serverUrn,
                                        AuthorityFinder.Purpose.REQUEST_RSPEC)
                        )
                        .collect(Collectors.toList());

                context.maxNodeNameLength = context.requestRspecSource.getBasicNodeInfo().stream()
                        .map(BasicStringRspec.BasicNodeInfo::getClientId)
                        .filter(Objects::nonNull)
                        .mapToInt(String::length)
                        .max()
                        .orElse(5); //uses 5 as fallback when no nodes with client_id found
            }

            ExperimentSetupConfig.ESpec eConfig = config.getESpec();
            if (eConfig != null) {
                assertNotNull(eConfig.getProvidedContentSource());

                context.eSpecBundle = createESpecBundle(eConfig, user);
                assertNotNull(context.eSpecBundle);

                logger.logExtraXml(ESpecBundle.ESPEC_DESCRIPTION_FILENAME + uniqueId,
                        context.eSpecBundle.getExperimentSpecificationYml(), true
                );

                ExperimentSpecificationParser parser = new ExperimentSpecificationParser();
                context.eSpec = parser.parse(context.eSpecBundle.getExperimentSpecificationYml());

                @SuppressWarnings("ConstantConditions")
                String userPem = user == null || user.getPrivateKey() == null ? null : new String(KeyUtil.privateKeyToAnyPem(user.getPrivateKey()));

                if (config.getOverrideESpecRSpec())  {
                    //Use RSpec from test config, which has already set:
                    //  context.requestRspec
                    //  context.requestRspecSource
                    //  context.serversInRspec
                    //  context.maxNodeNameLength
                    //  allRequestedNodeComponentIds

                    //overwrite in ESpec
                    ExperimentSpecificationBuilder newEspecBuilder = new BasicExperimentSpecificationBuilder(context.eSpec);
                    newEspecBuilder.setRspec(
                            new RspecSpec(
                                    new FileSource(DIRECT, context.requestRspec, "GuiLogicTest_generated_RSpec")
                            )
                    );
                    context.eSpec = newEspecBuilder.create();

                    context.eSpecBundle = new ModifiedESpecBundle(context.eSpecBundle, context.eSpec);
                } else {
                    //Get RSpec from ESpec
                    String requestRspec =
                            new String(new FileFetcher(
                                    context.eSpec.getRspecs().get(0).getSource(),
                                    context.eSpecBundle,
                                    rSpecGeneratorFactoryProvider.get(), new SingleSshGitAuthPreferences(userPem)).fetchBytes(),
                                    StandardCharsets.UTF_8);
                    context.requestRspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);

                    context.serversInRspec = context.requestRspecSource.getAllComponentManagerUrns().stream()
                            .map(serverUrn ->
                                    authorityFinder.findByUrn(
                                            serverUrn,
                                            AuthorityFinder.Purpose.REQUEST_RSPEC)
                            )
                            .collect(Collectors.toList());

                    context.maxNodeNameLength = context.requestRspecSource.getBasicNodeInfo().stream()
                            .map(BasicStringRspec.BasicNodeInfo::getClientId)
                            .filter(Objects::nonNull)
                            .mapToInt(String::length)
                            .max()
                            .orElse(5); //uses 5 as fallback when no nodes with client_id found

                    logger.logExtraXml("rspec" + uniqueId, requestRspec, true);

                    context.allRequestedNodeComponentIds.addAll(logRequestRspecFixedNodes(requestRspec));
                }
            }
        } catch (AdvertisementBasedRspecGenerator.RequiredResourcesUnavailableException e) {
            notEnoughFreeResourcesDetected = true;
            logger.fatal("Not enough free resources detected.", e);
        }
    }

    @Nonnull
    public static ESpecBundle createESpecBundle(@Nonnull ExperimentSetupConfig.ESpec eConfig, @Nullable GeniUser user)
            throws ESpecBundle.ESpecBundleInitException, InterruptedException {
        ESpecBundle res;
        switch (eConfig.getSource()) {
            case PROVIDE_ARCHIVE_FILE: {
                res = BundleFetcher.fetchArchiveFile(eConfig.getProvidedContentSource());
                break;
            }
            case PROVIDE_DIR: {
                res = BundleFetcher.fetchDir(eConfig.getProvidedContentSource());
                break;
            }
            case PROVIDE_ARCHIVE_URL: {
                res = BundleFetcher.fetchArchiveUrl(eConfig.getProvidedContentSource());
                break;
            }
            case PROVIDE_GIT_REPO_DIR: {
                //noinspection ConstantConditions
                String userPem =
                        user == null || user.getPrivateKey() == null ?
                                null :
                                new String(KeyUtil.privateKeyToAnyPem(user.getPrivateKey()));
                res =
                        BundleFetcher.fetchGitRepoDir(eConfig.getProvidedContentSource(), new SingleSshGitAuthPreferences(userPem));
                break;
            }
            default:
                throw new IllegalStateException("Unsupported espec.source: " + eConfig.getSource().name());
        }
        return res;
    }


    /** log any fixed nodes */
    @Nonnull
    private List<String> logRequestRspecFixedNodes(@Nullable String requestRspec) {
        try {
            if (requestRspec == null) return Collections.emptyList();

            RequestRspecSource requestRspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);
            List<BasicStringRspec.BasicNodeInfo> bni = requestRspecSource.getBasicNodeInfo();
            if (bni == null) return Collections.emptyList();

            List<String> fixedNodesList = bni.stream()
                    .filter(Objects::nonNull)
                    .map(BasicStringRspec.BasicNodeInfo::getComponentId)
                    .filter(Objects::nonNull)
                    .collect(Collectors.toList());
            if (fixedNodesList.isEmpty())
                return Collections.emptyList();

            logger.logExtraJson("rspec-request-fixed-nodes-" + uniqueId,
                    fixedNodesList, false);
            return Collections.unmodifiableList(fixedNodesList);
        } catch (Exception e) {
            logger.info("Exception while extracting fixed nodes from request rspec", e);
            return Collections.emptyList();
        }
    }

    private UserAndSliceApiWrapper userAndSliceApiWrapper;
    public void findProject() {
        assertNotNull(user);
        if (userAndSliceApiWrapper == null) {
            userAndSliceApiWrapper = auasawf.create();
        }

        switch (config.getSlice().getProjectSource()) {
            case NONE: {
                context.project = null;
                break;
            }
            case PROVIDED: {
                context.project = config.getSlice().getProject();
                break;
            }
            case AUTO: {
                List<String> subAuthorityNames;
                try {
                    if (!userAndSliceApiWrapper.hasUserCredentials()) {
                        userAndSliceApiWrapper.getUserCredentials(callLogger, user.getUserUrn());
                    }

                    subAuthorityNames =
                            userAndSliceApiWrapper.getSubAuthorityNames(callLogger, user.getUserUrn());
                } catch (Exception t) {
//                            LOG.warn("Exception while getting sub authority names. will try to ignore.", t);
                    logger.warn("Exception while getting sub authority names. will try to ignore. Exception: " + t.getMessage());
                    subAuthorityNames = Collections.emptyList();
                }

                if (!subAuthorityNames.isEmpty()) {
                    context.project = subAuthorityNames.get(0);
                } else {
                    UserAndSliceApiWrapper.SubAuthoritySupport subAuthoritySupport;
                    try {
                        subAuthoritySupport = userAndSliceApiWrapper.getSubAuthoritySupport(callLogger);
                    } catch (JFedException e) {
                        throw new RuntimeException("Failed to check if slice authority requires a mandatory authority/project for creating a slice");
                    }
                    if (Objects.equals(subAuthoritySupport, UserAndSliceApiWrapper.SubAuthoritySupport.SUB_AUTHORITY_MANDATORY)) {
                        throw new RuntimeException("There are no sub authorities for user " + user.getUserUrn());
                    } else {
                        //project prefered, but none found, so don't use one
                        context.project = null;
                    }
                }
                break;
            }
        }

        if (context.project != null) {
            logger.info("Will use project: \"" + context.project + "\"");
        } else {
            assertFalse(config.getSlice().getFailIfNoProject(), "Empty project is not allowed by config");
            logger.info("Will NOT use a project");
        }
    }

    @Nullable
    public Slice findExistingSlice(@Nonnull String sliceName, @Nullable String project, boolean checkSliceAuth) {
        final Server userAuthorityServer = user.getUserAuthorityServer();
        assert userAuthorityServer != null;
        String topLevelAuthority =
                userAuthorityServer.getDefaultComponentManagerAsGeniUrn().getEncodedTopLevelAuthority();
        if (context.project != null) {
            topLevelAuthority += ":" + context.project;
        }
        GeniUrn sliceUrn = GeniUrn.createGeniUrnFromEncodedParts(topLevelAuthority, "slice", context.sliceName);

        if (!checkSliceAuth) {
            return sfaModel.logExistSlice(sliceUrn);
        }

        if (!userAndSliceApiWrapper.hasUserCredentials()) {
            try {
                userAndSliceApiWrapper.getLocalUserCredentials(callLogger);
            } catch (JFedException e) {
                throw new RuntimeException("Failed to find existing slice: problem with user credential", e);
            }
        }

        try {
            List<AnyCredential> sliceCreds = userAndSliceApiWrapper.getSliceCredentials(callLogger, sliceUrn);
            if (sliceCreds == null || sliceCreds.isEmpty())
                return null;
        } catch (JFedException e) {
            logger.info("Could not get credentials for slice '"+sliceUrn+"'", e);
            return null;
        }

        return sfaModel.getSlice(sliceUrn);
    }

    //List<String> sshKeys = Collections.singletonList(nodeLoginTestStep.getSshKeyHelper().getSshPublicKeyString());

    public void createExperiment(String sliceName, List<UserSpec> userSpecs,
                                 boolean allowNewSlice, boolean allowRecoverSlice) {
        assertNotNull(user);
        context.sliceName = sliceName;

        logger.info("Will create slice \"" + context.sliceName + "\"");

        Instant requestedEndTime = Instant.now().plus(config.getSlice().getExpireTimeMin(), ChronoUnit.MINUTES);
        logger.info("Created slice (and slivers) will expire at " + requestedEndTime + "");

        if (context.eSpec == null) {
            assert context.requestRspecSource != null;
            assert context.requestRspecSource.getStringRspec() != null;
            assert context.requestRspecSource.getStringRspec().getXmlString() != null;
            assert !context.requestRspecSource.getStringRspec().getXmlString().trim().isEmpty();

            boolean createNew = allowNewSlice;
            Slice existingSlice = null;
            if (allowNewSlice && allowRecoverSlice) {
                //check if this is a new or a to be recovered slice
                existingSlice = findExistingSlice(context.sliceName, context.project, true);
                if (existingSlice != null)
                    createNew = false;
            }
            if (createNew) {
                context.experiment = new Experiment(
                        context.sliceName,
                        context.project,
                        Collections.emptyList(),
                        context.requestRspecSource,
                        null,
                        requestedEndTime,
                        userSpecs,
                        false,
                        config.getRunLinkTest());
            } else {
                //derive the expected sliceUrn
                if (existingSlice == null)
                    existingSlice = findExistingSlice(context.sliceName, context.project, false);

                if (existingSlice == null)
                    throw new RuntimeException("Implementation error: Failed to assume slice exists");

                //reset the slice to prevent old state from seeping through
                existingSlice = sfaModel.resetSlice(existingSlice.getUrn());

                context.experiment = new Experiment(
                        existingSlice,
                        context.requestRspecSource,
                        null,
                        requestedEndTime,
                        userSpecs,
                        null,
                        false,
                        config.getRunLinkTest());
            }
        } else {
            @SuppressWarnings("ConstantConditions")
            String userPem = user == null || user.getPrivateKey() == null ? null : new String(KeyUtil.privateKeyToAnyPem(user.getPrivateKey()));

            context.experiment = new Experiment(
                    context.sliceName,
                    context.project,
                    Collections.emptyList(),
                    context.eSpecBundle,
                    null,
                    requestedEndTime,
                    userSpecs,
                    false,
                    rSpecGeneratorFactoryProvider.get(),
                    new SingleSshGitAuthPreferences(userPem),
                    config.getRunLinkTest());
        }

        if (config.getWaitForReady().isEnabled()) {
            context.experiment.setMaxWaitUntilReady(
                    Duration.ofMinutes(config.getWaitForReady().getMaxTimeMin())
            );

            //The automated tester does not need to check each 10 seconds like the GUI
            context.experiment.setCheckReadyInterval(Duration.ofSeconds(30));

            if (context.experiment.getMaxWaitUntilReady() != null) {
                logger.info("Configured experiment for maximum wait time of " +
                        context.experiment.getMaxWaitUntilReady().toMinutes() + " minutes");
            }
        }

    }

    public void startExperiment(@Nonnull Injector injector) {
        assertTrue(config.getProvision().isEnabled(),
                "A disabled provision step is currently not supported.");

        context.notEnoughResourcesDetected = false;

        final ExperimentControllerFactory experimentControllerFactory =
                injector.getInstance(ExperimentControllerFactory.class);
        assertNotNull(experimentControllerFactory);

        context.experimentController =
                experimentControllerFactory.createExperimentController(context.experiment);

        context.experimentController.setExperimentLinkTesterFactory((experiment) -> new TestLinksJob(
                experiment,
                injector.getInstance(HighLevelTaskFactory.class),
                injector.getInstance(TaskThread.class),
                injector.getInstance(GeniUserProvider.class),
                injector.getInstance(JFedPreferences.class),
                injector.getInstance(ProxyPreferencesManager.class),
                injector.getInstance(ProxyServiceUtil.class),
                injector.getInstance(JobStateFactory.class),
                injector.getInstance(ProxySocketFactoryProvider.class))
        );

        assertNotNull(context.experimentController);


        context.experimentController.start();
        logger.info("Started experimentController. ExperimentState is now " + context.experiment.getExperimentState());

    }

    /**
     * returns true if the experiment is not ready or failed, and the deadline is not reached.
     *
     * Does not takes extraWaitTimeInSeconds into account!
     *    => once this returns false, the caller needs to start that wait.
     *
     * also sets context.waitForExperimentIsOver
     *       and context.notEnoughResourcesDetected
     */
    public boolean mustWaitForExperiment() {
        if (!context.waitForExperimentIsOver) {
            assertTrue(config.getWaitForReady().isEnabled(), "bug in runExperiment step code.");
            logger.info("Current ExperimentState=" + context.experiment.getExperimentState());
            logger.info("Checking if experiment is ready.");

            if (context.experiment.getSliceUrn() != null) {
                context.sliceUrn = context.experiment.getSliceUrn();
            }

            if (context.experiment.getExperimentState() == ExperimentState.READY) {
                logger.info("Ready, because ExperimentState is READY");
                context.waitForExperimentIsOver = true;
            }
            if (context.experiment.getAnyPartSawNotEnoughFreeResourcesError()) {
                context.waitForExperimentIsOver = true;
                context.notEnoughResourcesDetected = true;
                context.sawFailure = true;
                logger.info("An experiment failed: not enough free resources");
            } else {
                if (context.experiment.getExperimentState().isTerminated()) {
                    context.waitForExperimentIsOver = true;
                    context.sawFailure = true;
                    logger.info("Failure, because ExperimentState is terminated");
                }
                if (!context.experiment.getExperimentState().isActiveState()) {
                    context.waitForExperimentIsOver = true;
                    context.sawFailure = true;
                    logger.info("Failure, because ExperimentState is not active state");
                }
            }
        }
        return !context.waitForExperimentIsOver;
    }


    public boolean isNotEnoughFreeResourcesDetected() {
        return notEnoughFreeResourcesDetected;
    }

//assert handling

    @Contract("false -> fail")
    protected void assertTrue(boolean v) {
        assertTrue(v, "assertTrue failed");
    }

    @Contract("false,_ -> fail")
    protected void assertTrue(boolean v, @Nullable String failText) {
        if (!v)
            logger.fatal(failText);
    }

    @Contract("true -> fail")
    protected void assertFalse(boolean v) {
        assertFalse(v, "assertFalse failed");
    }

    @Contract("true,_ -> fail")
    protected void assertFalse(boolean v, @Nullable String failText) {
        if (v)
            logger.fatal(failText);
    }

    @Contract("null -> fail")
    protected void assertNotNull(@Nullable Object o) {
        assertNotNull(o, "object is null");
    }

    @Contract("null,_ -> fail")
    protected void assertNotNull(@Nullable Object o, @Nullable String failText) {
        if (o == null)
            logger.fatal(failText);
    }
}
